# -*- coding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    Copyright (C) 2004-2010 Tiny SPRL (<http://tiny.be>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import datetime
import operator
import string
import random

from osv import fields, osv
from tools.translate import _

TIME_UNIT_DAY = 1
TIME_UNIT_HOUR = 2
TIME_UNIT_DAY_LABEL = 'Day'
TIME_UNIT_HOUR_LABEL = 'Hour'
TIME_UNITS = [(TIME_UNIT_HOUR, TIME_UNIT_HOUR_LABEL), (TIME_UNIT_DAY, TIME_UNIT_DAY_LABEL)]
TIME_UNITS_DICT = dict(TIME_UNITS)

# Super-user identifier (aka Administrator aka root)
# TODO: Use the constant defined in osv.orm when OpenERP 6.1 is released
ROOT_USER_ID = 1

class hr_employee_sort_name(osv.osv):
    """Redefined HR employee to sort on name field."""
    # Need to be defined in the beginning of the file, otherwise
    # the module install process crashes with demo data...

    _inherit = "hr.employee"
    _order = "name_resource asc"

    _columns = {
        'name_resource': fields.related('resource_id', 'name', string="Name", type='char', size=128, store=True),
    }

hr_employee_sort_name()

class hrself_universe(osv.osv):
    """Group of employers."""

    _name = "hrself.universe"
    _description = "Group of employers"

    _columns = {
        # FIXME commented readonly to easy data insertion
        'name': fields.char('Name', size=48, required=True,),# readonly=True),
        'functional_administrator_id': fields.many2one('hr.employee', 'Master Functional Administrator', domain="[('id', 'in', functional_administrator_ids[0][2])]"),# readonly=True),
        'functional_administrator_ids': fields.many2many('hr.employee', 'univ_funct_admin_rel', 'universe_id', 'functional_administrator_id', 'Functional Administrators'),
        'groupes_id': fields.integer('Groupe S identifier'),
        'initialization_date': fields.datetime("Initialization date"),
    }

    def write(self, cr, uid, ids, values, context=None):
        """Redefined to set the user profile to human resource functional administrator"""
        if 'functional_administrator_ids' in values:
            functional_administrator_ids = values['functional_administrator_ids'][0]
            #For a many2many field, a list of tuples is expected.
            #(6, 0, [IDs]) replace the list of linked IDs (like using (5) then (4,ID) for each ID in the list of IDs --> unlink all --> link)
            if len(functional_administrator_ids) == 3 and functional_administrator_ids[0] == 6 and functional_administrator_ids[1] == 0:
                for employee_id in functional_administrator_ids[2]:
                    self.pool.get('hr.employee').browse(cr, uid, employee_id, context=context).user_id.set_admin_profile()
                for universe in self.browse(cr, uid, ids, context=context):
                    previous_admins = universe.functional_administrator_ids
                    if previous_admins:
                        for admin in previous_admins:
                            if admin.id not in functional_administrator_ids[2]:
                                # if master admin, can not delete it.
                                if admin.id == universe.functional_administrator_id.id:
                                    msg = 'You cannot delete master administrator !'
                                    raise osv.except_osv(_('Error !'), _(msg))
                                else:
                                    admin.user_id.del_admin_profile(context=context)

        return super(hrself_universe, self).write(cr, uid, ids, values, context)

    def fields_get(self, cr, uid, fields=None, context=None):
        """ Redefined as OpenERP does not handle fields access rules."""
        fields = super(hrself_universe, self).fields_get(cr, uid, fields, context)
        if not self.pool.get('res.users').is_in_admin_group(cr, uid, [uid], context):
            for field_name in ('name', 'initialization_date', 'employer_ids'):
                if field_name in fields:
                    fields[field_name]['readonly'] = True
        return fields

    def universe_id(self, cr, uid, context=None):
        """
        Returns the universe id, if it exists, corresponding to the user id.
        If the user is not linked to an employee, return the first universe id
        if there is one, false otherwise.

        :return: universe id or false
        :rtype: int or bool
        """
        employee_object = self.pool.get('hr.employee')
        employee_ids = employee_object.search(cr, uid, [('user_id', '=', uid)])
        if employee_ids:
            return employee_object.browse(cr, uid, employee_ids[0], context).universe_id.id
        else:
            universe_ids = self.search(cr, uid, [], context=context)
            universe_id = universe_ids and universe_ids[0] or False
            return universe_id

hrself_universe()

class hrself_level(osv.osv):
    """Level"""

    _name = "hrself.level"
    _description = "Level"

    def name_get(self, cr, uid, ids, context=None):
        """ Redefined to take into account the value field."""
        res = []
        for level in self.read(cr, uid, ids, ['name', 'value', 'code'], context):
            value_name = str(level['code']) + ' - ' + str(level['value'])
            res.append((level['id'], level['name'] or value_name))
        return res

    def name_search(self, cr, uid, name='', args=None, operator='ilike', context=None, limit=100):
        """ Redefined to take into account the value field."""
        result = super(hrself_level, self).name_search(cr, uid, name, args, operator, context, limit)
        if not result and name.isdigit():
            ids = self.search(cr, uid, [('value', '=', name)] + args, limit=limit, context=context)
            result = self.name_get(cr, uid, ids, context)
        return result

    def get_members(self, cr, uid, ids, context=None):
        """Returns the members (employees) of levels.
        :return: The members for each level.
        :rtype: A dictionary {level.id: [employee.id]}.
        (L{hrself_level}, L{hrself_employee})
        """
        result = dict.fromkeys(ids, [])
        services = self.pool.get('hrself.services').hrself_service(cr, uid, context=context)
        levels_members = services.get_levels_members(cr, uid, ids, datetime.date.today(), context=context)
        for i, level_members in enumerate(levels_members):
            result[ids[i]] = level_members
        return result

    def _get_members(self, cr, uid, ids, name, args, context=None):
        """Signature for function fields."""
        return self.get_members(cr, uid, ids, context=context)

    _columns = {
        'name': fields.char('Name', size=24, translate=True,),# readonly=True),
        'code': fields.selection([(1,'1'),(2,'2'),(3,'3')],'Level code',required=True),
        'value': fields.integer('Value', required=True),
        'member_ids': fields.function(_get_members, method=True, type='many2many', relation='hr.employee', string='Members'),
    }

hrself_level()

class hrself_employer(osv.osv):
    """Employer"""

    _name = "hrself.employer"
    _description = "Employer"

    _columns = {
        'name': fields.char('Name', size=48, required=True,),# readonly=True),
        'universe_id': fields.many2one('hrself.universe', 'Universe', required=True,),# readonly=True),
        'ccte_ids': fields.one2many('hrself.ccte', 'employer_id', 'CCTE'),
        'level_ids': fields.one2many('hrself.level', 'employer_id', 'Level'),
        'groupes_id': fields.integer('Groupe S identifier'),
    }

hrself_employer()

class hrself_level_extended(osv.osv):
    """Level"""

    _inherit = "hrself.level"

    _columns = {
        'employer_id': fields.many2one('hrself.employer', 'Employer', required=True),
    }

hrself_level_extended()

class hrself_level_combination(osv.osv):
    """Levels combination"""

    _name = "hrself.level.combination"
    _description = "Levels combination"

    def name_get(self, cr, uid, ids, context=None):
        """ Redefined as there is no name field."""
        res = []
        for level_combination in self.browse(cr, uid, ids, context):
            name = ' / '.join([level_combination[field].name_get(context=context)[0][1] for field in ['level_1_id', 'level_2_id', 'level_3_id'] if level_combination[field]])
            res.append((level_combination['id'], name))
        return res

    _columns = {
        'employer_id': fields.many2one('hrself.employer', 'Employer', required=True),
        'level_1_id': fields.many2one('hrself.level', 'Level 1', domain="[('employer_id', '=', employer_id), ('code', '=', 1)]"),
        'level_2_id': fields.many2one('hrself.level', 'Level 2', domain="[('employer_id', '=', employer_id), ('code', '=', 2)]"),
        'level_3_id': fields.many2one('hrself.level', 'Level 3', domain="[('employer_id', '=', employer_id), ('code', '=', 3)]"),
        'generate_date': fields.datetime('Generate date'),
    }

hrself_level_combination()

class hrself_universe_extended(osv.osv):

    _inherit = "hrself.universe"

    _columns = {
        'employer_ids': fields.one2many('hrself.employer', 'universe_id', 'Employers'),
    }

hrself_universe_extended()

class hrself_ccte(osv.osv):
    """Collective labour agreement"""

    _name = "hrself.ccte"
    _description = "Collective labour agreement"

    def name_get(self, cr, uid, ids, context=None):
        """ Redefined to take into account the groupes_id field."""
        res = []
        for ccte in self.read(cr, uid, ids, ['name', 'groupes_id'], context):
            res.append((ccte['id'], ccte['name'] or str(ccte['groupes_id'])))
        return res

    def name_search(self, cr, uid, name='', args=None, operator='ilike', context=None, limit=100):
        """ Redefined to take into account the groupes_id field."""
        result = super(hrself_ccte, self).name_search(cr, uid, name, args, operator, context, limit)
        if not result and name.isdigit():
            ids = self.search(cr, uid, [('groupes_id', '=', name)] + args, limit=limit, context=context)
            result = self.name_get(cr, uid, ids, context)
        return result

    _columns = {
        'name': fields.char('Name', size=24, translate=True,),# readonly=True),
        'employer_id': fields.many2one('hrself.employer', 'Employer', required=True,),# readonly=True),
        'groupes_id': fields.char('Groupe S identifier', size=6),
    }

hrself_ccte()

class hrself_code_list(osv.osv):
    """List of codes."""

    PERSONNEL_TYPE_NAME = 'CD_PERSONNEL' # Personnel type code name
    FULL_PARTIAL_TIME_NAME = 'CD_TEMPS_PARTIEL' # Full/partial time code name

    _name = "hrself.code.list"
    _description = "List of codes"

    _columns = {
        'groupes_id': fields.char('Groupe S identifier', size=30, required=True),
        'groupes_num_value': fields.integer('Groupe S numerical value', required=True),
        'name': fields.char('Name', size=50, translate=True, required=True),# readonly=True),
    }

hrself_code_list()

class hrself_activity_context(osv.osv):
    """Defines the context of an employee's activity."""

    # List of ordered non-function fields name subjected to uniqueness constraint
    UNIQUE_FIELDS_NAME = ['universe_id', 'employer_id', 'ccte_id', 'personnel_type_id', 'level_combination_id', 'full_partial_time_id']

    PERSONNEL_TYPE_DOMAIN = "[('groupes_id', '=', '%s')]" % hrself_code_list.PERSONNEL_TYPE_NAME # Personnel type domain
    FULL_PARTIAL_TIME_DOMAIN = "[('groupes_id', '=', '%s')]" % hrself_code_list.FULL_PARTIAL_TIME_NAME # Full/partial time domain

    _name = "hrself.activity.context"
    _description = "Activity context"

    def _get_order(self, cr, uid, ids, field_name, arg, context):
        """Returns for all activities contexts the corresponding order.
        @return: For all activities contexts in ids, the order
        @rtype dictionary where keys and values are integers
        """
        result = dict.fromkeys(ids, None)
        for activity_context in self.browse(cr, uid, ids, context):
            employer, ccte, personnel_type, level_combination, full_partial_time = \
                    operator.attrgetter('employer_id', 'ccte_id', 'personnel_type_id', 'level_combination_id', 'full_partial_time_id')(activity_context)
            if not employer: # (universe)
                order = 0
            elif not any((ccte, personnel_type, level_combination, full_partial_time)): # (universe, employer)
                order = 1
            elif ccte and not any((personnel_type, level_combination, full_partial_time)): # (universe, employer, ccte)
                order = 2
            elif personnel_type and not any((ccte, level_combination, full_partial_time)): # (universe, employer, personnel_type)
                order = 3
            elif level_combination and not any((ccte, personnel_type, full_partial_time)): # (universe, employer, level_combination)
                order = 4
            elif full_partial_time and not any((ccte, personnel_type, level_combination)): # (universe, employer, full_partial_time)
                order = 5
            else:
                order = 6
            result[activity_context.id] = order
        return result

    _columns = {
        'universe_id': fields.many2one('hrself.universe', 'Universe', required=True, select=True),
        'employer_id': fields.many2one('hrself.employer', 'Employer', select=True),
        'ccte_id': fields.many2one('hrself.ccte', 'CCTE', select=True),
        'personnel_type_id': fields.many2one('hrself.code.list', 'Personnel type', domain=PERSONNEL_TYPE_DOMAIN, select=True),
        'level_combination_id': fields.many2one('hrself.level.combination', 'Levels combination', select=True),
        'full_partial_time_id': fields.many2one('hrself.code.list', 'Full/Partial time code', domain=FULL_PARTIAL_TIME_DOMAIN, select=True),
        'order': fields.function(_get_order, method=True, type='integer'),
        'generate_date': fields.datetime('Generate date'),
    }

    def _default_universe(self, cr, uid, context=None):
        """Default universe."""
        return self.pool.get('hrself.universe').universe_id(cr, uid, context=context)

    _defaults = {
        'universe_id': _default_universe,
    }

    def name_get(self, cr, uid, ids, context=None):
        """ Redefined as there is no name field."""
        res = []
        for activity_context in self.browse(cr, uid, ids, context):
            name = ' / '.join([activity_context[field].name_get(context=context)[0][1] for field in self.UNIQUE_FIELDS_NAME if activity_context[field]])
            res.append((activity_context.id, name))
        return res

    _sql_constraints = [
        ('uniqueness', 'unique(%s)' % ', '.join(UNIQUE_FIELDS_NAME), _('Activity context must be unique')),
    ]

    def _check_unique_without_null(self, cr, uid, ids, context=None):
        """This method is used to complement the SQL UNIQUE constraint:
        according to the SQL standard, in the presence of a unique constraint it
        is possible to store duplicate rows that contain a null value in at
        least one of the constrained columns."""

        unique_fields = set(self.UNIQUE_FIELDS_NAME)
        for record in self.browse(cr, uid, ids):
            null_fields = set([field for field in unique_fields if isinstance(record[field], osv.orm.browse_null)])
            # if no field is null, then uniqueness has already been checked by the sql unique index
            # if some field(s) is/are null, we need to check by ourselves
            if len(null_fields) > 0:
                non_null_fields = unique_fields - null_fields
                domain = ['&'] * (len(unique_fields) - 1)
                domain.extend([(field, '=', None) for field in null_fields])
                domain.extend([(field, '=', record[field].id) for field in non_null_fields])
                if len(self.search(cr, uid, domain)) > 1: # result contains at least current activity context to be created
                    return False
        return True

    def _check_required_employer(self, cr, uid, ids, context=None):
        """Universe and employer are required if at least one of the other field is not null."""

        for context in self.browse(cr, uid, ids):
            if not context.employer_id and any((context.ccte_id, context.personnel_type_id, context.level_combination_id, context.full_partial_time_id)):
                return False
        return True

    _constraints = [
        (_check_unique_without_null, _('Activity context must be unique'), UNIQUE_FIELDS_NAME),
        (_check_required_employer, _('Employer is required'), ['employer_id']),
    ]

    def _is_field_compatible(self, some, other):
        """ Is some field compatible with other?
        @param some: a field
        @param other: another field
        @return: True if compatible, False otherwise
        @rtype: bool"""
        if (not (some or other)) or (other and not some) or (some and other and some.id == other.id):
            return True
        else:
            return False

    def is_compatible(self, cr, uid, some_ids, other_id):
        """ Is some_id compatible with other_id?
        @param some_id: identifier of an activity context L{hrself_activity_context}
        @type some_id: int
        @param other_id: identifier of an activity context L{hrself_activity_context}
        @type other_id: int
        @return: True if compatible, False otherwise
        @rtype: bool"""
        for some_context in self.browse(cr, uid, some_ids):
            other_context = self.browse(cr, uid, other_id)
            return all(
                [
                    self._is_field_compatible(some_context.universe_id, other_context.universe_id),
                    self._is_field_compatible(some_context.employer_id, other_context.employer_id),
                    self._is_field_compatible(some_context.ccte_id, other_context.ccte_id),
                    self._is_field_compatible(some_context.personnel_type_id, other_context.personnel_type_id),
                    self._is_field_compatible(some_context.level_combination_id, other_context.level_combination_id),
                    self._is_field_compatible(some_context.full_partial_time_id, other_context.full_partial_time_id),
                ]
            )

hrself_activity_context()

class hrself_prestation_code(osv.osv):
    """Defines the 'code prestation'."""

    _name = "hrself.prestation.code"
    _description = "Prestation code"

    def name_get(self, cr, uid, ids, context=None):
        """ Redefined to take into account the mnemonic and employer_id fields."""
        res = []
        for prestation_code in self.browse(cr, uid, ids, context):
            name = prestation_code.name or prestation_code.mnemonic
            if prestation_code.employer_id:
                name = '%s (%s)' % (name, prestation_code.employer_id.name_get(context=context)[0][1])
            res.append((prestation_code['id'], name))
        return res

    def name_search(self, cr, uid, name='', args=None, operator='ilike', context=None, limit=100):
        """ Redefined to take into account the mnemonic and employer_id fields."""
        result = super(hrself_prestation_code, self).name_search(cr, uid, name, args, operator, context, limit)
        if not result:
            ids = self.search(cr, uid, [('mnemonic', operator, name)] + args, limit=limit, context=context)
            result = self.name_get(cr, uid, ids, context)
        if not result:
            ids = self.search(cr, uid, [('employer_id', operator, name)] + args, limit=limit, context=context)
            result = self.name_get(cr, uid, ids, context)
        return result

    _columns = {
        'name': fields.char('Name', size=36, translate=True),
        'mnemonic': fields.char('Mnemonic', size=9, required=True),
        'employer_id': fields.many2one('hrself.employer', 'Employer'),
        'gross_leave': fields.boolean('Gross leave', help='Is this a gross leave event?'),
    }

hrself_prestation_code()

class hrself_common_type(osv.osv):
    """Features that are common between HRSelf types."""

    INIT_UNIT_YEAR_CIVIL = "0" #int 0 does not work as key for a selection field -> we use a str :(
    INIT_UNIT_YEAR_ONSS = "1"
    INIT_UNIT_QUARTER_CIVIL = "2"
    INIT_UNIT_QUARTER_ONSS = "3"
    INIT_UNIT_MONTH_CIVIL = "4"
    INIT_UNIT_MONTH_ONSS = "5"

    _name = "hrself.common.type"
    _description = "HRSelf common type"

    _columns = {
        'unit': fields.selection(TIME_UNITS, "Time unit", required=True),
        'has_initialization_value': fields.boolean('Has initialization value'),
        'init_value': fields.float("Initialization value"),
        'init_unit': fields.selection([
            (INIT_UNIT_YEAR_CIVIL, "Year (civil)"),
            (INIT_UNIT_YEAR_ONSS, "Year (ONSS)"),
            (INIT_UNIT_QUARTER_CIVIL, "Quarter (civil)"),
            (INIT_UNIT_QUARTER_ONSS, "Quarter (ONSS)"),
            (INIT_UNIT_MONTH_CIVIL, "Month (civil)"),
            (INIT_UNIT_MONTH_ONSS, "Month (ONSS)"),
            ], "Initialization unit"),
        'init_periodicity': fields.integer("Initialization periodicity"),
        'init_date': fields.date("Initialization date"),
    }

hrself_common_type()

class hrself_counter_type(osv.osv):
    """Defines the counter type."""

    _name = "hrself.counter.type"
    _inherit = "hrself.common.type"
    _description = "Counter type"

    _columns = {
        'name': fields.char('Name', size=24, translate=True),
        'mnemonic': fields.char('Mnemonic', size=9, required=True),
        'employer_id': fields.many2one('hrself.employer', 'Employer'),
    }

    def name_get(self, cr, uid, ids, context=None):
        """ Redefined to take into account the mnemonic and employer_id fields."""
        res = []
        for counter_type in self.browse(cr, uid, ids, context):
            name = counter_type.name or counter_type.mnemonic
            if counter_type.employer_id:
                name = '%s / %s' % (counter_type.employer_id.name_get(context=context)[0][1], name)
            res.append((counter_type['id'], name))
        return res

    def name_search(self, cr, uid, name='', args=None, operator='ilike', context=None, limit=100):
        """ Redefined to take into account the mnemonic and employer_id fields."""
        result = super(hrself_counter_type, self).name_search(cr, uid, name, args, operator, context, limit)
        if not result:
            names = name.split('/')
            if len(names) == 2:
                employer_name = names[0].strip()
                employer_ids = self.pool.get('hrself.employer').name_search(cr, uid, employer_name, args, operator, context, limit)
                mnemonic_name = names[1].strip()
                ids = self.search(cr, uid, [
                    ('employer_id', 'in', [employer_id for (employer_id, employer_name) in employer_ids]),
                    ('mnemonic', operator, mnemonic_name),
                ] + args, limit=limit, context=context)
            else:
                ids = self.search(cr, uid, [('mnemonic', operator, name)] + args, limit=limit, context=context)
                if not ids:
                    ids = self.search(cr, uid, [('employer_id', operator, name)] + args, limit=limit, context=context)
            result = self.name_get(cr, uid, ids, context)
        return result

hrself_counter_type()

class hr_employee_marital_status(osv.osv):

    _inherit = 'hr.employee.marital.status'

    _columns = {
        'name' : fields.char('Marital Status', size=30, required=True, translate=True),
        'groupes_id': fields.integer('Groupe S identifier'),
    }

hr_employee_marital_status()

class hr_employee(osv.osv):

    _inherit = "hr.employee"

    def name_get(self, cr, uid, ids, context=None):
        """ Redefined to take into account the first_name field."""
        res = []
        for employee in self.read(cr, uid, ids, ['name', 'first_name'], context):
            if employee['first_name']:
                name = '%s %s' % (employee['name'], employee['first_name'])
            else:
                name = employee['name']
            res.append((employee['id'], name))
        return res

    def name_search(self, cr, uid, name='', args=None, operator='ilike', context=None, limit=100):
        """ Redefined to take into account the first_name field."""
        result = super(hr_employee, self).name_search(cr, uid, name, args, operator, context, limit)
        names = name.split()
        if not result and len(names) == 2:
            ids = self.search(cr, uid, [('name', operator, names[0]), ('first_name', operator, names[1])] + args, limit=limit, context=context)
            result = self.name_get(cr, uid, ids, context)
        return result

    def _get_external(self, cr, uid, ids, field_name, arg, context):
        """Returns for all employees if there are external or not (i.e. if they have a Groupe S identifier).
        @return: For all employees in ids, True if external, False otherwise
        @rtype dictionary where keys are integers and values booleans
        """
        result = dict.fromkeys(ids, None)
        for employee in self.browse(cr, uid, ids):
            result[employee.id] = (employee.groupes_id == 0)
        return result

    def _get_hrself_email(self, cr, uid, ids, field_name, arg, context):
        """Returns for every employee the email used in HRSelf context,
        i.e. work email if defined, home email otherwise."""
        result = dict.fromkeys(ids, None)
        for employee in self.browse(cr, uid, ids, context=context):
            result[employee.id] = employee.work_email or employee.home_email
        return result

    def _get_photo(self, cr, uid, context=None):
        """ Redefined to change default employee's photo """
        photo_path = addons.get_module_resource('hr','images','hr-hover.png')
        return open(photo_path, 'rb').read().encode('base64')

    _columns = {
        'first_name': fields.char('First name', size=60),
        'universe_id': fields.many2one('hrself.universe', 'Universe'),
        'external': fields.function(_get_external, method=True, string='External', type='boolean'),
        'home_email': fields.related('address_home_id', 'email', type='char', size=240, string='Home E-mail'),
        'login': fields.related('user_id', 'login', type='char', string='Login'),
        'hrself_email': fields.function(_get_hrself_email, method=True, string='Email used in HRSelf context', type='char'),
        'groupes_id': fields.integer('Groupe S identifier'),
    }

    def _default_universe(self, cr, uid, context=None):
        """Default universe."""
        return self.pool.get('hrself.universe').universe_id(cr, uid, context=context)

    _defaults = {
        'universe_id': _default_universe,
    }

    def _check_universe_not_null(self, cr, uid, ids, context=None):
        """Check to see if universe_id is not null."""
        for employee in self.browse(cr, uid, ids, context=context):
            if not employee.external and not employee.universe_id:
                return False
        return True

    _constraints = [
        (_check_universe_not_null, _('Universe can not be null!'), ['universe_id']),
    ]

hr_employee()

class hrself_department(osv.osv):
    """Represents a department in a human resource context."""

    _inherit = 'hr.department'
    
    _columns = {
        'universe_id': fields.many2one('hrself.universe', 'Universe'),
        'level_id': fields.many2one('hrself.level','Corresponding level'),
        'level_member_ids': fields.related('level_id', 'member_ids', type='many2many', relation='hr.employee', readonly=True, string='Level members'),
    }    
    
    def _default_universe(self, cr, uid, context=None):
        """Default universe."""
        return self.pool.get('hrself.universe').universe_id(cr, uid, context=context)

    _defaults = {
        'universe_id': _default_universe,
    }

    def _check_universe_not_null(self, cr, uid, ids, context=None):
        """Check to see if universe_id is not null."""
        universe_ids = self.pool.get('hrself.universe').search(cr, uid, [], context=context)
        for department in self.browse(cr, uid, ids, context=context):
            if universe_ids and not department.universe_id:
                return False
        return True

    _constraints = [
        (_check_universe_not_null, _('Universe can not be null!'), ['universe_id']),
    ]

    def onchange_level_id(self, cr, uid, ids, name, level_id, context):
        """Action to execute when the user changes level_id."""
        result = {'value': {}}
        if level_id and not name:
            result['value']['name'] = self.pool.get('hrself.level').browse(cr, uid, level_id, context=context).name_get(context)[0][1]
        if level_id:
            members = self.pool.get('hrself.level').browse(cr, uid, level_id, context=context).member_ids
            result['value']['level_member_ids'] = [member.id for member in members]
        else:
            result['value']['level_member_ids'] = [(5)]
        return result

hrself_department()

class ir_model_data(osv.osv):
    _inherit = "ir.model.data"                                                                                                                  

    def get_resource_id(self, cr, uid, model, xml_id):
        """Get the resource identifier.
        :param model: model name
        :type model: str
        :param xml_id: xml identifier
        :type xml_id: str
        :return: resource identifier or False is resource not found
        :rtype: int or bool"""
        model_data_ids = self.search(cr, uid, [('model', '=', model), ('name', '=', xml_id)])
        if model_data_ids:
            return self.browse(cr, uid, model_data_ids[0]).res_id
        else:
            return False

ir_model_data()

class res_users(osv.osv):
    _inherit = "res.users"                                                                                                                  

    PROFILE_USER = 'user'
    PROFILE_APPROVER = 'approver'
    PROFILE_ADMIN = 'admin'

    def _get_employee(self, cr, uid, ids, fieldname, args, context=None):
        """Returns, for each user identifier, the associated employee identifier"""
        res = dict.fromkeys(ids, False)
        for user_id in ids:
            employee_ids = self.pool.get('hr.employee').search(cr, uid, [('user_id', '=', user_id)])
            if employee_ids:
                res[user_id] = employee_ids[0]
        return res

    def _store_employee(self, cr, uid, ids, context=None):
        """Returns the identifiers of the users whose employee_id field must be recomputed."""
        res = {}
        for employee in self.pool.get('hr.employee').browse(cr, uid, ids, context=context):
            res[employee.user_id.id] = True
        return res.keys()
    
    _columns = {
        'employee_id': fields.function(_get_employee, method=True, type='many2one', relation='hr.employee', store = {'hr.employee': (_store_employee, [], 10)}),
        # if first run, password must be changed by user
        'first_run': fields.boolean('First run', readonly=True),
    }

    _defaults = {
        'first_run': True,
    }

    def set_user_profile(self, cr, uid, ids, context=None):
        """Set the user profile to human resource user."""
        self._set_profile(cr, uid, ids, self.PROFILE_USER)

    def set_approver_profile(self, cr, uid, ids, context=None):
        """Set the user profile to human resource approver."""
        self._set_profile(cr, uid, ids, self.PROFILE_APPROVER)

    def set_admin_profile(self, cr, uid, ids, context=None):
        """Set the user profile to human resource functional administrator."""
        self._set_profile(cr, uid, ids, self.PROFILE_ADMIN)

    def _set_profile(self, cr, uid, ids, profile, context=None):
        """Set the user profile to ``profile``.

        WARNING: This really set the user profile: if the previous profile is
        admin and we set the profile to approver, then the user loses his
        admin profile!!! You have been warned..."""
        for user in self.browse(cr, uid, ids, context=context):
            old_groups = set([group_id.id for group_id in user.groups_id])
            new_groups = set()
            delete_groups = set()
            model_data_object = self.pool.get('ir.model.data')
            new_groups.add(model_data_object.get_resource_id(cr, uid, 'res.groups', 'group_hr_user'))
            approver_group_id = model_data_object.get_resource_id(cr, uid, 'res.groups', 'group_hr_approver')
            if profile == self.PROFILE_APPROVER or profile == self.PROFILE_ADMIN:
                new_groups.add(approver_group_id)
            else:
                delete_groups.add(approver_group_id)
            admin_group_id = model_data_object.get_resource_id(cr, uid, 'res.groups', 'group_hr_manager')
            if profile == self.PROFILE_ADMIN:
                new_groups.add(admin_group_id)
            else:
                delete_groups.add(admin_group_id)
            groups = (old_groups - delete_groups).union(new_groups)
            if groups ^ old_groups:
                #Use ROOT_USER_ID instead of uid to call the method with the admin privileges (à la sudo)
                self.write(cr, ROOT_USER_ID, ids, {
                    'groups_id': [(6, 0, [group_id for group_id in groups])],
                }, context=context) #For a many2many field: (6, 0, [IDs]) replace the list of linked IDs

    def del_admin_profile(self, cr, uid, ids, context=None):
        """Delete user's admin profile."""
        for user in self.browse(cr, uid, ids, context=context):
            if user.employee_id.is_approver():
                self.set_approver_profile(cr, uid, ids, context=context)
            else:
                self.set_user_profile(cr, uid, ids, context=context)

    def get_generated_login(self, name, identifier):
        """Returns the generated login associated with name and identifier.
        @rtype: string"""
        return name.lower().replace(' ', '') + str(identifier)

    def get_generated_password(self, cr, uid, ids, context=None):
        """Returns a generated password suitable for a user password.
        @rtype: string"""
        list = random.sample(string.ascii_lowercase, 6)
        list.append(random.choice(string.ascii_uppercase))
        list.append(random.choice(string.digits))
        random.shuffle(list)
        return ''.join(list)

    def _enforce_password_strength(self, cr, uid, ids, password, context=None):
        """Password must be 8 characters or more.
        Password must contain at least 3 of 4 types of characters: uppercase ascii, lowercase ascii, digit or punctuation symbol.
        Password must be different from login.
        @return True if password meets requirements.
        @rtype:boolean"""
        lowerc = upperc = digits = punctuation = 0
        if len(password) < 8:
            return False
        for res_user in self.browse(cr, uid, ids, context=context):
            if str(res_user.login).lower() == password.lower() :
                return False
        for letter in password:
            if letter in string.ascii_lowercase:
                lowerc = 1 
            elif letter in string.ascii_uppercase:
                upperc = 1
            elif letter in string.digits:
                digits = 1
            elif letter in string.punctuation:
                punctuation = 1
            else:
                return False
        if lowerc + upperc + digits + punctuation < 3:
            return False
        return True

    def is_in_admin_group(self, cr, uid, ids, context=None):
        """Returns True if at least one user is part of group_erp_manager (admin)."""
        return self._is_in_group(cr, uid, ids, 'group_erp_manager', context=context)

    def is_in_hr_admin_group(self, cr, uid, ids, context=None):
        """Returns True if at least one user is part of group_hr_manager 
        (Human Resources / Manager)."""
        return self._is_in_group(cr, uid, ids, 'group_hr_manager', context=context)

    def is_in_hr_approver_group(self, cr, uid, ids, context=None):
        """Returns True if at least one user is part of group_hr_approver 
        (Human Resources / Approver)."""
        return self._is_in_group(cr, uid, ids, 'group_hr_approver', context=context)

    def _is_in_group(self, cr, uid, ids, group, context=None):
        """Returns True if at least one user is part of group."""

        group_id = self.pool.get('ir.model.data').get_resource_id(cr, uid, 'res.groups', group)
        user_groups = set()
        for user in self.browse(cr, uid, ids, context=context):
            user_groups.update([group.id for group in user.groups_id])
        return group_id in user_groups
 
    def write(self, cr, uid, ids, values, context=None):
        """Redefined to enforce password strength."""
        if not hasattr(ids, '__iter__'):
            ids = [ids]
        password = values.get('password')
        if not password:
            password = values.get('clear_password') # XXX: depends on base_encrypt_password
        if password:
            if not self._enforce_password_strength(cr, uid, ids, password, context=context):
                msg = "The provided password is not strong enough. \n"\
                      "Password must contain at least 8 characters AND 3 of these 4 types of characters:\n"\
                      "1: Lowercase letter (abcde...z)\n"\
                      "2: Uppercase letter (ABCDE...Z)\n"\
                      "3: Digit (0123...9)\n"\
                      "4: Punctuation (%s)" % string.punctuation
                raise osv.except_osv(_('Error !'), _(msg))
        return super(res_users, self).write(cr, uid, ids, values, context=context)

res_users()

class EmailTemplateNotFound(Exception):
    pass

class hrself_mail(osv.osv):
    """Send an email."""

    _name = 'hrself.mail'

    def notify(self, cr, uid, email_template_name, ids):
        template_ids = self.pool.get('email.template').search(cr, uid, [('name', '=', email_template_name)])
        if not template_ids:
            raise EmailTemplateNotFound("Email template not found: %s" % email_template_name)
        return self.pool.get('email.template').generate_mail(cr, uid, template_ids[0], ids)

hrself_mail()

