# -*- coding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    Copyright (C) 2004-2010 Tiny SPRL (<http://tiny.be>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import datetime

from osv import osv
import hrself
import hrself_datetime
import logging

logger = logging.getLogger('hrself')

class ir_translation(osv.osv):
    """Translation."""

    _inherit = "ir.translation"

    def name_for_model_type(self, object_name, field_name):
        """Translation name field for a model type."""
        return '%s,%s' % (object_name, field_name)

ir_translation()

class hrself_manager(osv.osv_memory):
    """Facade to high level HRSelf services."""

    _name = "hrself.manager"
    _description = "Facade to high level hrself services"

    def create_hrself_code_list(self,cr,uid,values):
        """Create a HRSelf code list.

        :param values: Values
        :type values: dict
        :return: identifier of created HRSelf code list
        :rtype: integer"""
        return self._create_update_object_translation(cr, uid, None, 'hrself.code.list', values)

    def update_hrself_code_list(self,cr,uid,id,values):
        """Update a HRSelf code list.

        :param id: HRSelf code list identifier
        :type id: integer
        :param values: Values
        :type values: dict
        :return: True on sucess, False on failure
        :rtype: boolean"""
        result = self.pool.get('hrself.code.list').write(cr,uid,id,values)
        logger.info('hrself.code.list %d updated' % id)
        return result

    def read_hrself_code_list(self,cr,uid,identifier,value):
        """Read a HRSelf code list identified in Groupe S by (identifier, value).

        :param identifier: Groupe S code list identifier
        :type identifier: integer
        :param identifier: Groupe S code list value
        :type identifier: integer
        :return: Field values if HRSelf code list exists, False otherwise
        :rtype: dict or boolean"""
        code_list_object = self.pool.get('hrself.code.list')
        code_list_ids = code_list_object.search(cr,uid,[('groupes_id', '=', identifier), ('groupes_num_value', '=', value)])
        if code_list_ids:
            result = code_list_object.read(cr,uid,code_list_ids[0])
            result['name'] = self._read_translations(cr, uid, code_list_ids, 'hrself.code.list')
            return result
        else:
            return False

    def create_hrself_prestation_code(self, cr, uid, values):
        """Create a HRSelf prestation code.

        :param values: Values
        :type values: dict
        :return: identifier of created HRSelf prestation code
        :rtype: integer"""
        return self._create_update_object_translation(cr, uid, None, 'hrself.prestation.code', values)

    def update_hrself_prestation_code(self, cr, uid, id, values):
        """Update a HRSelf prestation code.

        :param id: HRSelf prestation code identifier
        :type id: integer
        :param values: Values
        :type values: dict
        :return: True on sucess, False on failure
        :rtype: boolean"""
        return self._create_update_object_translation(cr, uid, id, 'hrself.prestation.code', values)

    def read_hrself_prestation_code(self, cr, uid, employer_id, mnemonic):
        """Read a HRSelf prestation code identified in Groupe S by (employer_id, mnemonic).

        :param employer_id: Groupe S employer identifier
        :type employer_id: integer
        :param mnemonic: Groupe S mnemonic
        :type mnemonic: str
        :return: Field values if HRSelf prestation code exists, False otherwise
        :rtype: dict or boolean"""
        if employer_id:
            employer_object = self.pool.get('hrself.employer')
            employer_ids = employer_object.search(cr, uid, [('groupes_id', '=', employer_id)])
            employer_arg = ('employer_id', 'in', employer_ids)
        else:
            employer_ids = None
            employer_arg = ('employer_id', '=', None)
        prestation_code_object = self.pool.get('hrself.prestation.code')
        prestation_code_ids = prestation_code_object.search(cr, uid, [employer_arg, ('mnemonic', '=', mnemonic)])
        if prestation_code_ids:
            result = prestation_code_object.read(cr, uid, prestation_code_ids[0])
            if employer_ids:
                result['employer_id'] = employer_object.read(cr, uid, employer_ids[0])
            result['name'] = self._read_translations(cr, uid, prestation_code_ids, 'hrself.prestation.code')
            return result
        else:
            return False

    def create_hrself_counter_type(self, cr, uid, values):
        """Create a HRSelf counter type.

        :param values: Values
        :type values: dict
        :return: identifier of created HRSelf counter type
        :rtype: integer"""
        return self._create_update_object_translation(cr, uid, None, 'hrself.counter.type', self._hrself_counter_type_values(values))

    def update_hrself_counter_type(self, cr, uid, id, values):
        """Update a HRSelf counter type.

        :param id: HRSelf counter type identifier
        :type id: integer
        :param values: Values
        :type values: dict
        :return: True on sucess, False on failure
        :rtype: boolean"""
        return self._create_update_object_translation(cr, uid, id, 'hrself.counter.type', self._hrself_counter_type_values(values))

    def _hrself_counter_type_values(self, values):
        """OpenERP counter type values corresponding to Groupe S values.

        :param values: Groupe S values
        :type values: dict
        :return: OpenERP values
        :rtype: dict"""
        result = dict(values)
        if values.get('init_date'):
            result['init_date'] = hrself_datetime.to_date(values['init_date'].value, hrself_datetime.XML_RPC_DATETIME_FORMAT)
        return result

    def read_hrself_counter_type(self, cr, uid, employer_id, mnemonic):
        """Read a HRSelf counter type identified in Groupe S by (employer_id, mnemonic).
        
        :param employer_id: Groupe S employer identifier
        :type employer_id: integer
        :param mnemonic: Groupe S mnemonic
        :type mnemonic: str
        :return: Field values if HRSelf counter type exists, False otherwise
        :rtype: dict or boolean"""
        employer_object = self.pool.get('hrself.employer')
        employer_ids = employer_object.search(cr, uid, [('groupes_id', '=', employer_id)])
        counter_type_object = self.pool.get('hrself.counter.type')
        counter_type_ids = counter_type_object.search(cr, uid, [('employer_id', 'in', employer_ids), ('mnemonic', '=', mnemonic)])
        if counter_type_ids:
            result = counter_type_object.read(cr, uid, counter_type_ids[0])
            if employer_ids:
                result['employer_id'] = employer_object.read(cr, uid, employer_ids[0])
            result['name'] = self._read_translations(cr, uid, counter_type_ids, 'hrself.counter.type')
            if result.get('init_date'):
                result['init_date'] = hrself_datetime.to_datetime(result['init_date'], hrself_datetime.DATE_FORMAT)
            return result
        else:
            return False

    def create_hrself_universe(self,cr,uid,values):
        """Create a HRSelf universe.

        :param values: Values
        :type values: dict
        :return: identifier of created HRSelf universe
        :rtype: integer"""
        universe_object = self.pool.get('hrself.universe')
        universe_id = universe_object.create(cr,uid,values)
        logger.info('hrself.universe %d created' % universe_id)
        model_data_object = self.pool.get('ir.model.data')
        model_data_ids = model_data_object.search(cr,uid,[('module', '=', 'base'), ('name', '=', 'user_admin')])
        model_data = model_data_object.browse(cr,uid,model_data_ids[0])
        user_id = model_data.res_id
        user = self.pool.get('res.users').browse(cr,uid,user_id)
        employee_object = self.pool.get('hr.employee')
        employee_ids = employee_object.search(cr,uid,[('name', '=', user.name), ('user_id', '=', user.id)])
        if not employee_ids:
            employee_id = self._create_employee(cr, uid, {'universe_id': universe_id, 'name': user.name, 'user_id': user.id})
        else:
            employee_id = employee_ids[0]
            if not employee_object.browse(cr, uid, employee_id).universe_id:
                employee_object.write(cr, uid, employee_id, {'universe_id': universe_id})
        universe_object.write(cr, uid, [universe_id], {
            'functional_administrator_ids': [(6, 0, [employee_id])],
            'functional_administrator_id': employee_id,
        })
        company_object = self.pool.get('res.company')
        company_ids = company_object.search(cr, uid, [])
        if company_ids:
            company_object.write(cr, uid, company_ids, {'name': values['name']})
        return universe_id

    def update_hrself_universe(self,cr,uid,id,values):
        """Update a HRSelf universe.

        :param id: HRSelf universe identifier
        :type id: integer
        :param values: Values
        :type values: dict
        :return: True on sucess, False on failure
        :rtype: boolean"""
        result = self.pool.get('hrself.universe').write(cr,uid,id,values)
        logger.info('hrself.universe %d updated' % id)
        return result

    def read_hrself_universe(self,cr,uid,identifier):
        """Read a HRSelf universe identified in Groupe S by identifier.

        :param identifier: Groupe S universe identifier
        :type identifier: integer
        :return: Field values if HRSelf universe exists, False otherwise
        :rtype: dict or boolean"""
        universe_object = self.pool.get('hrself.universe')
        universe_ids = universe_object.search(cr,uid,[('groupes_id', '=', identifier)])
        if universe_ids:
            return universe_object.read(cr, uid, universe_ids[0], [
                'groupes_id',
                'name',
            ])
        else:
            return False

    def create_hrself_employer(self,cr,uid,values):
        """Create a HRSelf employer.

        :param values: Values
        :type values: dict
        :return: identifier of created HRSelf employer
        :rtype: integer"""
        employer_id = self.pool.get('hrself.employer').create(cr,uid,values)
        logger.info('hrself.employer %d created' % employer_id)
        return employer_id

    def update_hrself_employer(self, cr, uid, id, values):
        """Update a HRSelf employer.

        :param id: HRSelf employer identifier
        :type id: integer
        :param values: Values
        :type values: dict
        :return: True on sucess, False on failure
        :rtype: boolean"""
        result = self.pool.get('hrself.employer').write(cr, uid, id, values)
        logger.info('hrself.employer %d updated' % id)
        return result

    def read_hrself_employer(self,cr,uid,universe_identifier,employer_identifier):
        """Read a HRSelf employer identified in Groupe S by (universe_identifier, employer_identifier).

        :param universe_identifier: Groupe S universe identifier
        :type universe_identifier: integer
        :param employer_identifier: Groupe S employer identifier
        :type person_identifier: integer
        :return: Field values if HRSelf employer exists, False otherwise
        :rtype: dict or boolean"""
        universe_object = self.pool.get('hrself.universe')
        universe_ids = universe_object.search(cr,uid,[('groupes_id', '=', universe_identifier)])
        employer_object = self.pool.get('hrself.employer')
        employer_ids = employer_object.search(cr,uid,[('universe_id', 'in', universe_ids), ('groupes_id', '=', employer_identifier)])
        if employer_ids:
            result = employer_object.read(cr,uid,employer_ids[0])
            result['universe_id'] = universe_object.read(cr,uid,universe_ids[0])
            return result
        else:
            return False

    def create_hrself_ccte(self,cr,uid,values):
        """Create a HRSelf CCTE.

        :param values: Values
        :type values: dict
        :return: identifier of created HRSelf CCTE
        :rtype: integer"""
        return self._create_update_object_translation(cr, uid, None, 'hrself.ccte', values)

    def update_hrself_ccte(self,cr,uid,id,values):
        """Update a HRSelf CCTE.

        :param id: HRSelf CCTE identifier
        :type id: integer
        :param values: Values
        :type values: dict
        :return: True on sucess, False on failure
        :rtype: boolean"""
        pass

    def read_hrself_ccte(self,cr,uid,employer_identifier,ccte_identifier):
        """Read a HRSelf CCTE identified in Groupe S by (employer_identifier, ccte_identifier).

        :param employer_identifier: Groupe S employer identifier
        :type employer_identifier: integer
        :param ccte_identifier: Groupe S CCTE identifier
        :type ccte_identifier: str
        :return: Field values if HRSelf CCTE exists, False otherwise
        :rtype: dict or boolean"""
        employer_object = self.pool.get('hrself.employer')
        employer_ids = employer_object.search(cr,uid,[('groupes_id', '=', employer_identifier)])
        ccte_object = self.pool.get('hrself.ccte')
        ccte_ids = ccte_object.search(cr,uid,[('employer_id', 'in', employer_ids), ('groupes_id', '=', ccte_identifier)])
        if ccte_ids:
            result = ccte_object.read(cr,uid,ccte_ids[0])
            result['employer_id'] = employer_object.read(cr,uid,employer_ids[0])
            result['name'] = self._read_translations(cr, uid, ccte_ids, 'hrself.ccte')
            return result
        else:
            return False

    def create_hrself_level(self, cr, uid, values):
        """Create a HRSelf level.

        :param values: Values
        :type values: dict
        :return: identifier of created HRSelf level
        :rtype: integer"""
        return self._create_update_object_translation(cr, uid, None, 'hrself.level', values)

    def update_hrself_level(self, cr, uid, id, values):
        """Update a HRSelf level.

        :param id: HRSelf level identifier
        :type id: integer
        :param values: Values
        :type values: dict
        :return: True on sucess, False on failure
        :rtype: boolean"""
        return self._create_update_object_translation(cr, uid, id, 'hrself.level', values)

    def read_hrself_level(self,cr,uid,employer_identifier,level_code, level_value):
        """Read a HRSelf level identified in Groupe S by (employer_identifier, level_code, level_value).

        :param employer_identifier: Groupe S employer identifier
        :type employer_identifier: integer
        :param level_code: Groupe S level code
        :type level_code: integer
        :param level_value: Groupe S level value
        :type level_value: integer
        :return: Field values if HRSelf level exists, False otherwise
        :rtype: dict or boolean"""
        employer_object = self.pool.get('hrself.employer')
        employer_ids = employer_object.search(cr,uid,[('groupes_id', '=', employer_identifier)])
        level_object = self.pool.get('hrself.level')
        level_ids = level_object.search(cr,uid,[('employer_id', 'in', employer_ids), ('code', '=', level_code), ('value', '=', level_value)])
        if level_ids:
            result = level_object.read(cr, uid, level_ids[0], ['name', 'code', 'value'])
            result['employer_id'] = employer_object.read(cr,uid,employer_ids[0])
            result['name'] = self._read_translations(cr, uid, level_ids, 'hrself.level')
            return result
        else:
            return False

    def create_hrself_person(self, cr, uid, values):
        """Create a HRSelf person.

        :param values: Values
        :type values: dict
        :return: identifier of created HRSelf person
        :rtype: integer"""
        user_id = self._create_user(cr, uid, values)
        values['user_id'] = user_id
        employee_id = self._create_employee(cr, uid, self._hrself_person_values(cr, uid, values=values))
        return employee_id

    def _create_user(self, cr, uid, values):
        """Create a user.

        :return: identifier of created user
        :rtype: integer"""
        user_object = self.pool.get('res.users')
        name = values['name']
        user_values = {
            'name': name, 
            'login': user_object.get_generated_login(name, values['groupes_id']), 
            'active': values['active'],
        }
        if values.get('context_lang'):
            user_values['context_lang'] = values['context_lang']
        user_id = user_object.create(cr, uid, user_values)
        logger.info('res.users %d created' % user_id)
        user_object.browse(cr, uid, user_id).set_user_profile()
        return user_id

    def _create_employee(self, cr, uid, values):
        """Create an employee.

        :return: identifier of created employee
        :rtype: integer"""
        employee_id = self.pool.get('hr.employee').create(cr, uid, values)
        logger.info('hr.employee %d created' % employee_id)
        return employee_id

    def update_hrself_person(self, cr, uid, id, values):
        """Update a HRSelf person.

        :param id: HRSelf person identifier
        :type id: integer
        :param values: Values
        :type values: dict
        :return: True on sucess, False on failure
        :rtype: boolean"""
        user_id = self.pool.get('hr.employee').browse(cr, uid, id).user_id.id
        result = self.pool.get('res.users').write(cr, uid, user_id, {'active': values['active']})
        logger.info('res.users %d updated' % user_id)
        #Backward compatible with previous version of replication
        values.pop('context_lang', None)
        result = self.pool.get('hr.employee').write(cr, uid, id, self._hrself_person_values(cr, uid, id, values))
        logger.info('hr.employee %d updated' % id)
        return result

    def _hrself_person_values(self, cr, uid, employee_id=0, values=None):
        """OpenERP person values corresponding to Groupe S values.

        :param employee_id: employee identifier, if any
        :type employee_id: int
        :param values: Groupe S values
        :type values: dict
        :return: OpenERP values
        :rtype: dict"""
        if values is None:
            values = {}
        result = dict(values)
        if values.get('country_id'):
            country_ids = self.pool.get('res.country').search(cr, uid, [('code', '=', values['country_id'])])
            if country_ids:
                result['country_id'] = country_ids[0]
            else:
                del result['country_id']
                logger.warning('No country for code %s' % (values['country_id']))
        if values.get('birthday'):
            result['birthday'] = hrself_datetime.to_date(values['birthday'].value, hrself_datetime.XML_RPC_DATETIME_FORMAT)
        if values.get('marital'):
            marital = values.pop('marital')
            if 'name' in marital:
                names = marital.pop('name')
                if len(names):
                    source_name = names['en_US']
                    marital['name'] = source_name
                    marital_status_object = self.pool.get('hr.employee.marital.status')
                    marital_status_ids = marital_status_object.search(cr, uid, [('name', '=', source_name)])
                    if marital_status_ids:
                        marital_status_id = marital_status_ids[0]
                        if not marital_status_object.browse(cr, uid, marital_status_id).groupes_id:
                            marital_status_object.write(cr, uid, marital_status_id, marital)
                            logger.info('hr.employee.marital.status %d updated' % marital_status_id)
                    else:
                        marital_status_id = self.pool.get('hr.employee.marital.status').create(cr, uid, marital)
                        logger.info('hr.employee.marital.status %d created' % marital_status_id)
                    result['marital'] = marital_status_id
                    self._create_update_translations(cr, uid, marital_status_id, 'hr.employee.marital.status', source_name, names)
        address_object = self.pool.get('res.partner.address')
        employee = employee_id and self.pool.get('hr.employee').browse(cr, uid, employee_id)
        if 'work_email' in values:
            if not values['work_email']:
                address_id = False
            elif employee and employee.address_id and employee.address_id.email == values['work_email']:
                address_id = employee.address_id.id
            else:
                address_id = address_object.create(cr, uid, {'email': values['work_email']})
            result['address_id'] = address_id
        if 'home_email' in values:
            if not values['home_email']:
                address_id = False
            elif employee and employee.address_home_id and employee.address_home_id.email == values['home_email']:
                address_id = employee.address_home_id.id
            else:
                address_id = address_object.create(cr, uid, {'email': values['home_email']})
            result['address_home_id'] = address_id
        return result

    def search_hrself_person(self, cr, uid, universe_identifier, person_identifier):
        """Search a HRSelf person identified in Groupe S by (universe_identifier, person_identifier).

        :param universe_identifier: Groupe S universe identifier
        :type universe_identifier: integer
        :param person_identifier: Groupe S person identifier
        :type person_identifier: integer
        :return: OpenERP Identifier if the person exists, False otherwise
        :rtype: integer or boolean"""
        universe_ids = self.pool.get('hrself.universe').search(cr, uid, [('groupes_id', '=', universe_identifier)])
        employee_ids = self.pool.get('hr.employee').search(cr, uid, [('universe_id', 'in', universe_ids), ('groupes_id', '=', person_identifier)])
        if employee_ids:
            return employee_ids[0]
        else:
            return False

    def read_hrself_person(self, cr, uid, id):
        """Read a HRSelf person identified by id.

        :precondition: the person identified by id exists
        :return: Field values of the HRSelf person
        :rtype: dict"""
        result = self.pool.get('hr.employee').read(cr, uid, id, [
            'universe_id', 'user_id', 'marital', 'country_id', 'birthday', 'groupes_id', 
            'name', 'first_name', 'gender', 'active', 'work_email', 'home_email',
        ])
        result['universe_id'] = self.pool.get('hrself.universe').read(cr, uid, result['universe_id'][0])
        result['user_id'] = self.pool.get('res.users').read(cr, uid, result['user_id'][0])
        if result['marital']:
            result['marital'] = self.pool.get('hr.employee.marital.status').read(cr, uid, result['marital'][0])
        if result['country_id']:
            result['country_id'] = self.pool.get('res.country').read(cr, uid, result['country_id'][0])
        if result['birthday']:
            result['birthday'] = hrself_datetime.to_datetime(result['birthday'], hrself_datetime.DATE_FORMAT)
        return result

    def _level_id(self, cr, uid, employer_id, code, value):
        """HRSelf level identifier corresponding to Groupe S (code, value) level."""
        level_ids = self.pool.get('hrself.level').search(cr, uid, [('employer_id', '=', employer_id), ('code', '=', code), ('value', '=', value)])
        if level_ids:
            return level_ids[0]

    def create_hrself_activity_context(self, cr, uid, values):
        """Create a HRSelf activity context if it doesn't exists.

        :param values: Values
        :type values: dict
        :return: identifier of HRSelf activity context (created or existing)
        :rtype: integer"""
        universe_id = self.pool.get('hrself.universe').search(cr,uid,[('groupes_id', '=', values['universe_id'])])[0]
        employer_id = self.pool.get('hrself.employer').search(cr, uid, [('universe_id', '=', universe_id), ('groupes_id', '=', values['employer_id'])])[0]
        ccte_id = self.pool.get('hrself.ccte').search(cr, uid, [('employer_id', '=', employer_id), ('groupes_id', '=', values['ccte_id'])])[0]
        personnel_type_id = self.pool.get('hrself.code.list').search(cr, uid, [('groupes_id', '=', 'CD_PERSONNEL'), ('groupes_num_value', '=', values['personnel_type_id'])])[0]
        level_combination = values['level_combination_id']
        level_combination_values = {
            'employer_id': employer_id, 
            'level_1_id': self._level_id(cr, uid, employer_id, 1, level_combination['level_1_id']),
            'level_2_id': self._level_id(cr, uid, employer_id, 2, level_combination['level_2_id']),
            'level_3_id': self._level_id(cr, uid, employer_id, 3, level_combination['level_3_id']),
        }
        level_combination_object = self.pool.get('hrself.level.combination')
        level_combination_ids = level_combination_object.search(cr, uid, [(key, '=', value) for key, value in level_combination_values.items()])
        if level_combination_ids:
            level_combination_id = level_combination_ids[0]
        else:
            level_combination_values['generate_date'] = datetime.datetime.today()
            level_combination_id = level_combination_object.create(cr, hrself.ROOT_USER_ID, level_combination_values)
            logger.info('hrself.level.combination %d created' % level_combination_id)
        full_partial_time_id = self.pool.get('hrself.code.list').search(cr, uid, [('groupes_id', '=', 'CD_TEMPS_PARTIEL'), ('groupes_num_value', '=', values['full_partial_time_id'])])[0]
        values = {
            'universe_id': universe_id,
            'employer_id': employer_id,
            'ccte_id': ccte_id,
            'personnel_type_id': personnel_type_id,
            'level_combination_id': level_combination_id,
            'full_partial_time_id': full_partial_time_id,
        }
        activity_context_object = self.pool.get('hrself.activity.context')
        activity_context_ids = activity_context_object.search(cr, uid, [(key, '=', value) for key, value in values.items()])
        if activity_context_ids:
            activity_context_id = activity_context_ids[0]
        else:
            values['generate_date'] = datetime.datetime.today()
            activity_context_id = activity_context_object.create(cr, hrself.ROOT_USER_ID, values)
            logger.info('hrself.activity.context %d created' % activity_context_id)
        return activity_context_id

    def read_hrself_activity_context(self, cr, uid, values):
        """Read a HRSelf activity context based on values.

        :param values: values
        :type values: dict
        :return: Field values if HRSelf activity context exists, False otherwise
        :rtype: dict or boolean"""
        universe_object = self.pool.get('hrself.universe')
        universe_ids = universe_object.search(cr,uid,[('groupes_id', '=', values['universe_id'])])
        employer_object = self.pool.get('hrself.employer')
        employer_ids = employer_object.search(cr,uid,[('universe_id', 'in', universe_ids), ('groupes_id', '=', values['employer_id'])])
        if employer_ids:
            ccte_object = self.pool.get('hrself.ccte')
            ccte_ids = ccte_object.search(cr, uid, [('employer_id', 'in', employer_ids), ('groupes_id', '=', values['ccte_id'])])
            code_list_object = self.pool.get('hrself.code.list')
            personnel_type_ids = code_list_object.search(cr, uid, [('groupes_id', '=', 'CD_PERSONNEL'), ('groupes_num_value', '=', values['personnel_type_id'])])
            level_combination = values['level_combination_id']
            level_combination_values = {
                'employer_id': employer_ids[0], 
                'level_1_id': self._level_id(cr, uid, employer_ids[0], 1, level_combination['level_1_id']),
                'level_2_id': self._level_id(cr, uid, employer_ids[0], 2, level_combination['level_2_id']),
                'level_3_id': self._level_id(cr, uid, employer_ids[0], 3, level_combination['level_3_id']),
            }
            level_combination_object = self.pool.get('hrself.level.combination')
            level_combination_ids = level_combination_object.search(cr, uid, [(key, '=', value) for key, value in level_combination_values.items()])
            full_partial_time_ids = code_list_object.search(cr, uid, [('groupes_id', '=', 'CD_TEMPS_PARTIEL'), ('groupes_num_value', '=', values['full_partial_time_id'])])
            if all([universe_ids, employer_ids, ccte_ids, personnel_type_ids, level_combination_ids, full_partial_time_ids]):
                activity_context_object = self.pool.get('hrself.activity.context')
                activity_context_ids = activity_context_object.search(cr, uid, [
                    ('universe_id', '=', universe_ids[0]),
                    ('employer_id', '=', employer_ids[0]),
                    ('ccte_id', '=', ccte_ids[0]),
                    ('personnel_type_id', '=', personnel_type_ids[0]),
                    ('level_combination_id', '=', level_combination_ids[0]),
                    ('full_partial_time_id', '=', full_partial_time_ids[0]),
                    ])
                if activity_context_ids:
                    return activity_context_object.read(cr, uid, activity_context_ids[0])
        return False

    def _create_update_object_translation(self, cr, uid, id, model_name, values):
        """Create or update a HRSelf object that has a translatable field.

        :param id: 
            - create: identifier of HRSelf object
            - update: None
        :type id: int
        :param model_name: the model name
        :type model_name: str
        :param values: Values
        :type values: dict
        :return: 
            - create: identifier of created HRSelf object
            - update: True on success, False on failure
        :rtype: 
            - create: integer
            - update: bool"""
        names = values.pop('name')
        source_name = names.get('en_US')
        if source_name:
            values['name'] = source_name
        if id:
            result = self.pool.get(model_name).write(cr, uid, id, values)
            logger.info('%s %d updated' % (model_name, id))
        else:
            result = id = self.pool.get(model_name).create(cr, uid, values)
            logger.info('%s %d created' % (model_name, id))
        if names:
            self._create_update_translations(cr, uid, id, model_name, source_name, names)
        return result

    def _create_update_translations(self, cr, uid, id, model_name, source_name, names):
        """Create or update the translations associated to a model.

        :param id: the model identifier
        :type id: int
        :param model_name: the model name
        :type model_name: str
        :param source_name: the source name for the translation
        :type source_name: str
        :param names: the translations
        :type names: dict where the keys are the language and the value the translation in that language"""
        translation_object = self.pool.get('ir.translation')
        for (key, value) in names.iteritems():
            if value:
                translation_values = {'lang': key, 'src': source_name, 'name': translation_object.name_for_model_type(model_name, 'name'), 'type': 'model', 'res_id': id}
                translation_ids = translation_object.search(cr, uid, [(tkey, '=', tvalue) for tkey, tvalue in translation_values.items()])
                if translation_ids:
                    translation_id = translation_ids[0]
                    if value != translation_object.browse(cr, uid, translation_id).value:
                        translation_object.write(cr, uid, [translation_id], {'value': value})
                        logger.info('ir.translation %d updated' % translation_id)
                else:
                    translation_values['value'] = value
                    translation_id = translation_object.create(cr, uid, translation_values)
                    logger.info('ir.translation %d created' % translation_id)

    def _read_translations(self, cr, uid, ids, model_name):
        """Read the translations associated to a model.

        :param ids: the model identifiers
        :type ids: list of integer
        :param model_name: the model name
        :type model_name: str
        :return: the translations
        :rtype: dict where the keys are the language and the value the translation in that language"""
        translations = {}
        translation_object = self.pool.get('ir.translation')
        translation_ids = translation_object.search(cr,uid,[('name', '=', translation_object.name_for_model_type(model_name, 'name')),('type', '=', 'model'),('res_id', 'in', ids)])
        for translation in translation_object.browse(cr,uid,translation_ids):
            translations[translation.lang] = translation.value
        return translations

hrself_manager()
