# -*- coding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    Copyright (C) 2004-2010 Tiny SPRL (<http://tiny.be>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from __future__ import with_statement
import logging
import xmlrpclib
import socket
import pprint

import tools
from osv import osv
from tools.translate import _
import hrself_datetime

logger = logging.getLogger('hrself.services')

def log(title, message, channel=logging.DEBUG, depth=None):
    """Log a message."""
    logger = logging.getLogger(title)
    if logger.isEnabledFor(channel):
        for line in pprint.pformat(message, depth=depth).split('\n'):
            logger.log(channel, line)

class HRSelfContextManager(object):
    """HRSelf context manager."""

    def __init__(self, cr, uri, root, context=None):
        """Constructor.
        @param: uri: URI of the proxy
        @param root: root path"""
        self._message = None
        self.cr = cr
        self.proxy = xmlrpclib.ServerProxy(uri)
        self._root = root
        self.localcontext = context

    def __enter__(self):
        """Enter the proxy."""
        return self

    def is_connected(self):
        """Is the proxy connected?"""
        proxy_root = getattr(self.proxy, self._root)
        if proxy_root.est_connecte():
            return True
        if proxy_root.connexion_perdue():
            proxy_root.reconnecter()
        if not proxy_root.est_connecte():
            self._message = _('%s is not connected.') % str(self.proxy)
            return False
        else:
            return True

    def __exit__(self, exception_type, exception_value, exception_traceback):
        """Exit the proxy."""
        if exception_type == socket.error:
            self._message = _('Unable to connect using url %s') % str(self.proxy)
        elif exception_type in (xmlrpclib.ResponseError, xmlrpclib.Fault):
            self._message = '%s: %s' % (str(self.proxy), exception_value)
        if self._message:
            logger.warning(self._message)
            raise osv.except_osv(_('Error !'), _( """This operation is currently not available.
                          Please retry later or contact your functional administrator."""))

class hrself_services(osv.osv_memory):
    """HRSelf services factory."""

    _name = 'hrself.services'

    def __init__(self, pool, cr):
        """Constructor."""
        self._cache = {}
        super(hrself_services, self).__init__(pool, cr)

    def _cache_key(self, proxy_service, dbname):
        """Format used for self._cache key"""
        return '%s-%s' % (proxy_service, dbname)

    def _service(self, cr, uid, proxy_service, stub_service, universe_id=None, context=None):
        """Returns the service associated to the given arguments."""
        if 'hrself_server_url' in tools.config.options:
            if self._cache_key(proxy_service, cr.dbname) in self._cache:
                return self._cache[self._cache_key(proxy_service, cr.dbname)]
            else:
                url = tools.config['hrself_server_url']
                universe_object = self.pool.get('hrself.universe')
                if universe_id is None:
                    universe_id = universe_object.search(cr, uid, [])[0]
                universe = universe_object.browse(cr, uid, universe_id)
                employer_ids = universe.employer_ids
                if employer_ids:
                    employer_id = employer_ids[0].groupes_id
                    with HRSelfContextManager(cr, url, 'hrself_localisateur', context=context) as context_manager:
                        if context_manager.is_connected():
                            if context_manager.proxy.hrself_localisateur.existe_employeur(employer_id):
                                url = context_manager.proxy.hrself_localisateur.serveur_pour_employeur(employer_id)
                                if url:
                                    result = proxy_service(url, self.pool)
                                    self._cache[self._cache_key(proxy_service, cr.dbname)] = result
                                    return result
                                else:
                                    raise osv.except_osv(_('Error !'), _("""No server available for the employer identified by %d!""") % employer_id)
                            else:
                                raise osv.except_osv(_('Error !'), _("""The employer identified by %d doesn't exist!""") % employer_id)
                else:
                    raise osv.except_osv(_('Error !'), _("""The universe identified by %d has no employers!""") % employee.universe_id.groupes_id.id)
        else:
            message = _("Missing option: '%s', using a stub instead") % 'hrself_server_url' 
            logger.warning(message)
            return stub_service("stub", self.pool)

    def hrself_service(self, cr, uid, universe_id=None, context=None):
        """Returns the HRSelf service associated to the given arguments."""
        return self._service(cr, uid, ProxyService, ServiceStub, universe_id=universe_id, context=context)
        
hrself_services()

class Service(object):
    """Interface to services needed by the HRSelf project."""

    def __init__(self, url, pool):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.url = url
        self.pool = pool

    def get_levels_members(self, cr, uid, ids, date, universe_id=None, context=None):
        """Returns for every level the members of this level.
        :param ids: level identifiers
        :type ids: list of int or int
        :param date: date
        :return: for every level, the members of each level.
        :rtype: list or list of list of integer
        """
        raise NotImplementedError()

class ServiceStub(Service):
    """Simulates proxy service for decoupled testing.
    We do not simulate an xmlrpc server so we do not test for xmlrpc exception.
    We do not simulate activity context create_or_update so we don't call hrself.manager."""

    def get_levels_members(self, cr, uid, ids, date, universe_id=None, context=None):
        """Returns for every level the members of this level.
        :param ids: level identifiers
        :type ids: list of int or int
        :param date: date
        :return: for every level, the members of each level.
        :rtype: list or list of list of integer
        """
        return []

class ProxyService(Service):
    """Service implemented through a proxy."""

    def __init__(self, url, pool):
        """Constructor."""
        super(ProxyService, self).__init__(url, pool)
        self.url = url
        self.pool = pool

    def get_levels_members(self, cr, uid, ids, date, universe_id=None, context=None):
        """Returns for every level the members of this level.
        :param ids: level identifiers
        :type ids: list of int or int
        :param date: date
        :return: for every level, the members of each level.
        :rtype: list or list of list of integer
        """
        result = []
        universe_object = self.pool.get('hrself.universe')
        if universe_id is None:
            universe_id = universe_object.search(cr, uid, [])[0]
        universe_external_id = universe_object.browse(cr, uid, universe_id).groupes_id
        datetime = hrself_datetime.to_datetime(date)
        levels = []
        if isinstance(ids, int):
            level_ids = [ids]
        else:
            level_ids = ids
        for level in self.pool.get('hrself.level').browse(cr, uid, level_ids):
            levels.append({
                'code': level.code,
                'value': level.value,
            })
        with HRSelfContextManager(cr, self.url, 'hrself', context=context) as context_manager:
            if context_manager.is_connected():
                levels_members = context_manager.proxy.hrself.personnes_niveaux(universe_external_id, levels, datetime)
                for level_members in levels_members:
                    employee_ids = self.pool.get('hr.employee').search(cr, uid, [('groupes_id', 'in', level_members)])
                    result.append(employee_ids)
        if isinstance(ids, int) and len(result) > 0:
            result = result[0]
        log('levels_members', result)
        return result
