# -*- coding: utf-8 -*-
##############################################################################
#    
#    OpenERP, Open Source Enterprise Management Solution
#    Copyright (C) 2004-2010 Tiny SPRL (<http://openerp.com>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.     
#
##############################################################################

import datetime
from osv import fields, osv
from tools.translate import _

class hrself_password_wizard(osv.osv_memory):
    """Wizard to regenerate the password of selected users 
    and send them an email with the new password."""

    _name = 'hrself.password.wizard'

    def _default_employees(self, cr, uid, ids, context=None):
        """Default employees, i.e. the one that have no passwords."""
        user_ids = self.pool.get('res.users').search(cr, uid, [
            '|', ('password', '=', ''), 
            ('password', '=', False),
        ], context=context)
        employee_ids = self.pool.get('hr.employee').search(cr, uid, [('user_id', 'in', user_ids)], context=context)
        return employee_ids

    _columns = {
        'employee_id': fields.many2one('hr.employee', 'Employee', required=True),
        'login': fields.related('employee_id', 'user_id', 'login', type='char', size=64, string='Login', required=True),
        'password': fields.char('Password', size=64),
        'random_password': fields.boolean('Random Password'),
        'employee_ids': fields.many2many('hr.employee', 'wizard_employee_rel', 'wizard_id', 'employee_id', 'Employee', readonly=True),
    }

    _defaults = {
        'employee_ids': _default_employees,
    }

    def notify_new_password(self, cr, uid, ids, context=None):
        """Sends an email to the employee the password has been regenerated."""
        email_template_name = '_notify_employee_of_new_password'
        hrself_mail = self.pool.get('hrself.mail')
        hrself_mail.notify(cr, uid, email_template_name, ids)

    def generate_new(self, cr, uid, ids, context=None):
        """Generate the password of new users, i.e. the one that have no passwords."""
        wizard = self.browse(cr, uid, ids, context=context)[0]
        employee_ids = [employee.id for employee in wizard.employee_ids]
        if employee_ids:
            context['active_ids'] = employee_ids
            return self.regenerate(cr, uid, ids, context=context)
        else:
            return {'type': 'ir.actions.act_window_close'}

    def regenerate(self, cr, uid, ids, context=None):
        """Regenerate the password of selected users."""
        if context is None:
            context = {}
        login_ids = []
        model_data_object = self.pool.get('ir.model.data')
        employee_ids = context.get('active_ids') or [context.get('active_id')]

        user_ids = [employee.user_id.id for employee in self.pool.get('hr.employee').browse(cr, uid, employee_ids, context)]
        if self.pool.get("res.users").is_in_admin_group(cr, uid, user_ids, context):
            msg = "Please unselect any administrator from the list. Administrators can change their password in the preferences menu."
            raise osv.except_osv(_('Error !'), _(msg))
        if uid in user_ids:
            msg = "Please unselect yourself from the list. Use the preferences menu to change your password."
            raise osv.except_osv(_('Error !'), _(msg))
        for employee in self.pool.get('hr.employee').browse(cr, uid, employee_ids, context=context):
            w = self.browse(cr, uid, ids, context=context)
            password = w[0].random_password and employee.user_id.get_generated_password() or w[0].password
            employee.user_id.write({
                'password': password,
                'clear_password': password,
            }, context=context)
            login_id = self.create(cr, uid, {
                'employee_id': employee.id, 
                'password': password,
            }, context=context)
            if employee.hrself_email:
                self.notify_new_password(cr, uid, [login_id], context=context)
                self.unlink(cr, uid, [login_id], context=context)
            else:
                login_ids.append(login_id)
        if login_ids:
            model_data_id = model_data_object._get_id(cr, uid, 'hrself', 'hrself_login_wizard_action')
            window_id = model_data_object.browse(cr, uid, model_data_id, context=context).res_id
            result = self.pool.get('ir.actions.act_window').read(cr, uid, window_id, context=context)
            result['domain'] = "[('id', 'in', %s)]" % login_ids
            return result
        else:
            return {'type': 'ir.actions.act_window_close'}
        
hrself_password_wizard()

class hrself_annual_initialization_wizard(osv.osv_memory):
    """Wizard for the HRSelf annual initialization."""

    _name = 'hrself.annual.initialization.wizard'

    def initialize(self, cr, uid, ids, context=None):
        """Do the initialization."""
        universe = self.pool.get('res.users').browse(cr, uid, uid, context=context).employee_id.universe_id
        if not universe.initialization_date:
            return self._initialize(cr, uid, ids, context)
        else:
            model_data_object = self.pool.get('ir.model.data')
            model_data_id = model_data_object._get_id(cr, uid, 'hrself', 'hrself_annual_initialization_wizard_confirmation_action')
            window_id = model_data_object.browse(cr, uid, model_data_id, context=context).res_id
            result = self.pool.get('ir.actions.act_window').read(cr, uid, window_id, context=context)
            return result

    def reinitialize(self, cr, uid, ids, context=None):
        """Redo the initialization."""
        return self._initialize(cr, uid, ids, context)

    def _initialize(self, cr, uid, ids, context=None):
        """Private implementation of the initialization."""
        return {}

hrself_annual_initialization_wizard()
