# -*- coding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    Copyright (C) 2004-2010 Tiny SPRL (<http://tiny.be>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import datetime
import urllib
from hrself import hrself_datetime
from osv import fields, osv
import tools

class hr_employee(osv.osv):
    """Represents an employee in a human resource context."""

    _inherit = 'hr.employee'

    def _get_current_leave_approver(self, cr, uid, ids, name, args, context=None):
        """Returns the current leave approver (the one that is defined for today) for an employee.
        @return: The current leave approver for each employee.
        @rtype: A dictionary where keys are employee ids and values are approver ids
        """
        res = dict.fromkeys(ids, False)
        for employee in self.browse(cr, uid, ids, context=context):
            approvement_object = self.pool.get('hrself.employee.approvement')
            approvement_ids = approvement_object.search(cr, uid, [('employee_id', '=', employee.id)])
            for approvement in approvement_object.browse(cr, uid, approvement_ids):
                if datetime.date.today() in hrself_datetime.period(approvement.start_date, approvement.end_date):
                    res[employee.id] = approvement.approver_id.id
        return res

    def _get_approver_from_definition(self, cr, uid, ids, context=None):
        """Get the leave approver from the approvement definition."""
        res = []
        for approvement in self.pool.get('hrself.employee.approvement').browse(cr, uid, ids, context=context):
            res.append(approvement.employee_id.id)
        return res
    
    def _get_backup_from_definition(self, cr, uid, ids, context=None):
        """Get the leave backup from the approvement definition."""
        res = []
        for approvement in self.pool.get('hrself.backup.approvement').browse(cr, uid, ids, context=context):
            res.append(approvement.approver_id.id)
        return res
    
    def _get_current_leave_backup(self, cr, uid, ids, name, args, context=None):
        """Returns the current leave backup (the one that is defined for today) for an approver.
        @return: The current leave backup for each approver.
        @rtype: A dictionary of tuple where the tuple is 
        (approver id, approver backup id)
        (L{hrself_employee}, L{hrself_employee})
        """
        res = dict.fromkeys(ids, False)
        for approver in self.browse(cr, uid, ids, context=context):
            approvement_object = self.pool.get('hrself.backup.approvement')
            approvement_ids = approvement_object.search(cr, uid, [('approver_id', '=', approver.id)])
            for approvement in approvement_object.browse(cr, uid, approvement_ids):
                if datetime.date.today() in hrself_datetime.period(approvement.start_date, approvement.end_date):
                    res[approver.id] = approvement.backup_approver_id.id
        return res
    
    def _get_actual_leave_approver(self, cr, uid, ids, name, args, context=None):
        """Returns the actual leave approver, i.e. (s)he is not absent or on leave and (s)he can approve right now."""

        today = datetime.date.today()
        return self.get_leave_approver(cr, uid, ids, today, context)

    def _get_openerp_server_url(self, cr, uid, ids, field_name, arg, context=None):
        """Returns url of openerp server.
        """
        result = dict.fromkeys(ids)
        if 'openerp_server_url' in tools.config.options:
            base_url = tools.config['openerp_server_url']
            for employee in self.browse(cr, uid, ids):
                login = employee.user_id.login
                universe = employee.universe_id.name
                url = base_url + '?universe=' + urllib.quote(universe.encode('utf8')) + '&user=' + urllib.quote(login.encode('utf8'))
                result[employee.id] = url
        return result

    def _get_current_department(self, cr, uid, ids, field_name, arg, context=None):
        """Returns the current department of employee.
            If employee is attached to multiple departments (by level), only one is returned.
        """
        result = dict.fromkeys(ids, None)
        for employee in self.browse(cr, uid, ids, context=context):
            if employee.department_id:
                result[employee.id] = employee.department_id.id
            else:
                #if department defined by level
                department_object = self.pool.get('hr.department')
                universe_id = employee.universe_id.id
                # check first if department attached to a level exists
                universe_department_ids = department_object.search(cr, uid, [('universe_id', '=', universe_id), ('level_id', '!=', None)], context = context)
                if universe_department_ids:
                    #get levels of employee and check department attached
                    activity_context_id = employee.get_today_activity_context_id(context=context)
                    activity_context_object = self.pool.get('hrself.activity.context')
                    activity_context = activity_context_object.browse(cr, uid, activity_context_id, context=context)
                    level_combination = activity_context.level_combination_id
                    employee_levels = [level_combination.level_1_id.id, level_combination.level_2_id.id, level_combination.level_3_id.id]
                    department_ids = department_object.search(cr, uid, [('id', 'in', universe_department_ids), ('level_id', 'in', employee_levels)], context = context)
                    for department in department_object.browse(cr, uid, department_ids, context=context):
                        result[employee.id] = department.id
        return result

    _columns = {
        'leave_approver_id': fields.function(
                                _get_current_leave_approver,
                                method=True,
                                type='many2one',
                                relation='hr.employee',
                                string='Current leave approver',
                                store={ 'hrself.employee.approvement': (_get_approver_from_definition, ['employee_id', 'approver_id', 'start_date', 'end_date'], 10) },
                            ),
        # backup is defined for an employee that is himself an approver
        'leave_backup_id': fields.function(
                                _get_current_leave_backup,
                                method=True,
                                type='many2one',
                                relation='hr.employee',
                                string='Current leave backup',
                                store={ 'hrself.backup.approvement': (_get_backup_from_definition, ['backup_approver_id', 'approver_id', 'start_date', 'end_date'], 10) },
                            ),
        # actual leave approver provided is not absent or on leave
        'actual_leave_approver_id': fields.function(_get_actual_leave_approver, method=True, type='many2one', relation='hr.employee', string='Actual leave approver'),
        'employee_approvement_ids': fields.one2many('hrself.employee.approvement', 'employee_id', string='Employee Approvement'),
        'employee_approvement_backup_ids': fields.one2many('hrself.backup.approvement', 'approver_id', string='Employee Back up'),
        'workflow_id': fields.many2one('workflow', 'Request Leave Approvement Workflow'),
        'openerp_server_url': fields.function(_get_openerp_server_url, method=True, type='str'),
        'current_department_id': fields.function(_get_current_department, method=True, string='Department', type='many2one', relation='hr.department'),
    }

    def is_approver(self, cr, uid, ids, context=None):
        """Check if employee is an approver (undepending of corresponding user's groups!)
            If employee is backup only, not considered approver.
        """
        result = False
        employee_approvement_object = self.pool.get('hrself.employee.approvement')
        department_approvement_object = self.pool.get('hrself.department.approvement')
        for employee in self.browse(cr, uid, ids, context=context):
            if employee_approvement_object.search(cr, uid, [('approver_id', '=', employee.id )], context=context) or department_approvement_object.search(cr, uid, [('approver_id', '=', employee.id) ], context=context) :
                result = True
        return result

    def get_approvees(self, cr, uid, ids, context=None):
        """Returns list of employees the current employee is direct approver AND employees from departments the current employee is approver. (undepending approvement absence)
        @param ids: ids of employees
        @rtype: A dictionary {employee id: [list of employee's ids]}
        """
        result = dict.fromkeys(ids,[])
        employees_list = []
        for employee in self.browse(cr, uid, ids, context):
            employees_list.extend(employee.get_direct_approvees(context=context)[employee.id])
            employees_list.extend(employee.get_department_approvees(context=context)[employee.id])
            result[employee.id] = employees_list 
        return result

    def get_direct_approvees(self, cr, uid, ids, context=None):
        """Returns list of employees the current employee is direct approver. (undepending approvement absence)
        @param ids: ids of employees
        @rtype: A dictionary {employee id: [list of employee's ids]}
        """
        result = dict.fromkeys(ids,[])
        for employee in self.browse(cr, uid, ids, context):
            if employee.is_approver(context=context):
                employee_approvement_object = self.pool.get('hrself.employee.approvement')
                approvement_ids = employee_approvement_object.search(cr, uid, [('approver_id', '=', employee.id )], context=context)
                employees_list = []
                for emp_approvement in employee_approvement_object.browse(cr, uid, approvement_ids, context=context):
                    if datetime.date.today() in hrself_datetime.period(emp_approvement.start_date, emp_approvement.end_date):
                        employees_list.append(emp_approvement.employee_id.id)
                result[employee.id] = employees_list 
        return result

    def get_department_approvees(self, cr, uid, ids, context=None):
        """Returns list of employees of all departments the current employee is approver. (undepending approvement absence)
        @param ids: ids of employees
        @rtype: A dictionary {employee id: [list of employee's ids]}
        """
        result = dict.fromkeys(ids,[])
        for employee in self.browse(cr, uid, ids, context):
            if employee.is_approver(context=context):
                department_approvement_object = self.pool.get('hrself.department.approvement')
                approvement_ids = department_approvement_object.search(cr, uid, [('approver_id', '=', employee.id )], context=context)
                for dep_approvement in department_approvement_object.browse(cr, uid, approvement_ids, context=context):
                    if datetime.date.today() in hrself_datetime.period(dep_approvement.start_date, dep_approvement.end_date):
                        department_id = dep_approvement.department_id.id
                        department_object = self.pool.get('hr.department')
                        department = department_object.browse(cr, uid, department_id, context=context)
                        members_list = []
                        for member in department.member_ids:
                            members_list.append(member.id)
                        result[employee.id] = members_list 
        return result

    def _is_absent_for_leave_approvement(self, cr, uid, ids, date_from, date_to=None, context=None):
        """Returns True if the employee is set as absent for leave approvement or on leave during the specified period, False otherwise.
        @rtype: boolean."""

        is_absent = self.pool.get('hrself.holidays.request.absence')._is_absent_for_leave_approvement(cr, uid, ids, date_from, date_to, context) 
        is_on_leave = self.pool.get('hrself.holidays.request.leave')._is_on_leave(cr, uid, ids, date_from, date_to, context)
        return is_absent or is_on_leave

    def _get_assigned_approver(self, cr, uid, ids, date, context=None):
        """Returns the direct assigned leave approver for a specific day or None if none found."""
        result = dict.fromkeys(ids)
        for employee in self.browse(cr, uid, ids, context):
            if employee.leave_approver_id:
                if self._is_absent_for_leave_approvement(cr, uid, [employee.leave_approver_id.id], date, date, context):
                    if employee.leave_approver_id.leave_backup_id:
                        # if approver is absent, approver backup cannot be absent
                        result[employee.id] = employee.leave_approver_id.leave_backup_id.id
                    else:
                        result[employee.id] = employee.universe_id.functional_administrator_id.id
                else:
                    result[employee.id] = employee.leave_approver_id.id
        return result

    def _get_direct_department_approver(self, cr, uid, ids, date, context=None):
        """Returns the direct department leave approver for a specific day or None if none found."""
        result = dict.fromkeys(ids)
        for employee in self.browse(cr, uid, ids, context=context):
            department = employee.department_id
            if department:
                result[employee.id] = department.get_leave_approver(date, context=context)[department.id]
        return result

    def _get_department_approver(self, cr, uid, ids, date, context=None):
        """Returns the department leave approver for a specific day. If no direct department approver is found,
        then returns the approver of the department implicitly referenced by the level of the theoretical activity,
        or None if not found."""
        if context is None:
            context = {}
        direct_approvers = self._get_direct_department_approver(cr, uid, ids, date, context)
        approvers = dict(direct_approvers)
        employee_ids = [employee_id for (employee_id, approver_id) in direct_approvers.items() if approver_id is None]
        if len(employee_ids) > 0:
            level_ids = []
            universe_id = self.pool.get('hr.employee').browse(cr, uid, employee_ids[0]).universe_id.id
            department_osv = self.pool.get('hr.department')
            department_ids = department_osv.search(cr, uid, [('universe_id', '=', universe_id), ('level_id', '!=', None)])
            approver_ids = department_osv.get_leave_approver(cr, uid, department_ids, date, context=context)
            for department in department_osv.browse(cr, uid, department_ids, context=context):
                if approver_ids[department.id]:
                    level_ids.append(department.level_id.id)
            levels_members = self.pool.get('hrself.level').get_members(cr, uid, level_ids, context=context)
            for (level_id, member_ids) in levels_members.items():
                department_ids = department_osv.search(cr, uid, [('universe_id', '=', universe_id), ('level_id', '=', level_id)])
                for employee_id in employee_ids:
                    if employee_id in member_ids:
                        approvers[employee_id] = approver_ids[department_ids[0]]

        result = dict(approvers)
        for (employee_id, approver_id) in approvers.items():
            if approver_id == employee_id:
                result[employee_id] = None
        return result

    def get_leave_approver(self, cr, uid, ids, date, context=None):
        """Returns the direct leave approver for a specific day or None if none found."""
        result = {}
        if len(ids) > 0:
            assigned_approvers = self._get_assigned_approver(cr, uid, ids, date, context=context)
            employee_ids = [employee_id for (employee_id, approver_id) in assigned_approvers.items() if approver_id is None]
            department_approvers = self._get_department_approver(cr, uid, employee_ids, date, context=context)
            result = dict(assigned_approvers, **department_approvers)
            employee_ids = [employee_id for (employee_id, approver_id) in department_approvers.items() if approver_id is None]
            if len(employee_ids) > 0:
                functional_administrator_id = self.browse(cr, uid, ids[0], context=context).universe_id.functional_administrator_id.id
                admin_approvers = dict.fromkeys(employee_ids, functional_administrator_id)
                result.update(admin_approvers)
        return result
    
    def get_activity_context_id(self, cr, uid, ids, date_from, date_to, context=None):
        """Returns the activity context for a specific day or None if none found.
        @param: ids = id of employee
        @return list of activity context id
        @rtype: list
        """
        #activity_context_id = {}
        #TODO: must return a list, not an id
        services = self.pool.get('hrself.services').hrself_holidays_service(cr, uid, context=context)
        for employee in self.browse(cr, uid, ids, context=context):
            activities = services.get_theoretical_activities(cr, uid, employee.id, date_from, date_to, context=context)
            if activities:
                activity_context_id = activities[0]['activity_context']
        return activity_context_id

    def get_one_day_activity_context_id(self, cr, uid, ids, date, context=None):
        """Returns the activity context for a specific day or None if none found.
        @param: date = specific day of activity context
        @param: ids = id of employee
        @return list of activity context id
        @rtype: list
        """
        return self.get_activity_context_id(cr, uid, ids, date_from=date, date_to=date, context=context)

    def get_today_activity_context_id(self, cr, uid, ids, context=None):
        """Returns the activity context for a specific day or None if none found.
        @param: ids = id of employee
        @return list of activity context id
        @rtype: list
        """
        today = datetime.date.today()
        return self.get_one_day_activity_context_id(cr, uid, ids, date=today , context=context)

hr_employee()

