# -*- coding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    Copyright (C) 2004-2010 Tiny SPRL (<http://tiny.be>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import datetime
from osv import fields,osv
from hrself import hrself_datetime
from tools.translate import _

class hrself_approvement(osv.osv):
    """Approvement definition.
    This is just an abstract osv to be used by concrete ones that are defined below.
    TODO: avoid creating associated table.
    """

    _name = "hrself.approvement"

    _columns = {
        'approver_id': fields.many2one('hr.employee', 'Approver', required=True),
        'start_date': fields.date('Start date', required=True),
        'end_date': fields.date('End date'),
    }

    _defaults = {
        'start_date': hrself_datetime.to_string(datetime.date.today())
    }

    def _set_approver_profile(self, cr, uid, employee_id, context=None):
        """Set the user (linked to employee) profile to Human Resources / Approver."""
        user = self.pool.get('hr.employee').browse(cr, uid, employee_id, context=context).user_id
        # If the user is already in HR approver group, we do nothing: indeed, 
        # if the user is in HR manager group (and hence in HR approver group), 
        # then he would leave the HR manager group, which is not what we want!
        if not user.is_in_hr_approver_group():
            user.set_approver_profile()

    def create(self, cr, uid, values, context=None):
        """Redefined to set the user profile to human resource approver"""
        if 'approver_id' in values:
            self._set_approver_profile(cr, uid, values['approver_id'], context=context)
        return super(hrself_approvement, self).create(cr, uid, values, context)

    def write(self, cr, uid, ids, values, context=None):
        """Redefined to set the user profile to human resource approver"""
        if 'approver_id' in values:
            self._set_approver_profile(cr, uid, values['approver_id'], context=context)
        return super(hrself_approvement, self).write(cr, uid, ids, values, context)

    def _check_ordered_dates(self, cr, uid, ids, context=None):
        """ Check if dates are time ordered """
        for approvement in self.browse(cr, uid, ids, context=context):
            start_date = approvement.start_date
            end_date = approvement.end_date
        if end_date and end_date < start_date:
            return False
        return True

    def _check_period(self, cr, uid, ids, context=None):
        """ Check time period compatibility with other approvement definitions """
        for approvement in self.browse(cr, uid, ids, context=context):
            start_date = approvement.start_date
            end_date = approvement.end_date
        if end_date:
            self_period = hrself_datetime.period(start_date, end_date)
        else:
            self_period = hrself_datetime.period(start_date)
        approvement_ids = self._get_approvement_ids(cr, uid, ids, context=context)
        for approvement in self.browse(cr, uid, approvement_ids, context=context):
            if approvement.end_date:
                approvement_period = hrself_datetime.period(approvement.start_date, approvement.end_date)
            else:
                approvement_period = hrself_datetime.period(approvement.start_date)
            if self_period.overlap(approvement_period):
                return False
        return True

    def _get_approvement_ids(self, cr, uid, ids, context=None):
        """ Get approvement ids """
        raise NotImplementedError()

    _constraints = [
        (_check_ordered_dates, _('End date must be greater than or equal to start date'), ['end_date']),
        (_check_period, _('Period of approvement overlaps an existing one.'), ['end_date']),
    ]


hrself_approvement()

class hrself_employee_approvement(osv.osv):
    """Employee approvement definition."""

    _name = "hrself.employee.approvement"
    _inherit = "hrself.approvement"
    _description = "Employee approvement definition"

    _columns = {
        'employee_id': fields.many2one('hr.employee', 'Employee', required=True),
    }

    def _get_approvement_ids(self, cr, uid, ids, context=None):
        """ Get approvement ids. Every approvement definition for selected employee except current one """
        for approvement in self.browse(cr, uid, ids, context=context):
            employee_id = approvement.employee_id.id
        approvement_ids = self.pool.get('hrself.employee.approvement').search(cr, uid, ['&', ('id','!=',ids), ('employee_id','=',employee_id), ], context=context)
        return approvement_ids

hrself_employee_approvement()

class hrself_backup_approvement(osv.osv):
    """Backup approvement definition."""

    _name = "hrself.backup.approvement"
    _inherit = "hrself.approvement"
    _description = "Backup approvement definition"

    _columns = {
        'backup_approver_id': fields.many2one('hr.employee', 'Backup approver', required=True),
    }

    def create(self, cr, uid, values, context=None):
        """Redefined to set the user profile to human resource approver"""
        if 'backup_approver_id' in values:
            self._set_approver_profile(cr, uid, values['backup_approver_id'], context=context)
        return super(hrself_backup_approvement, self).create(cr, uid, values, context)

    def write(self, cr, uid, ids, values, context=None):
        """Redefined to set the user profile to human resource leave approver"""
        if 'backup_approver_id' in values:
            self._set_approver_profile(cr, uid, values['backup_approver_id'], context=context)
        return super(hrself_backup_approvement, self).write(cr, uid, ids, values, context)

    def _get_approvement_ids(self, cr, uid, ids, context=None):
        """ Get approvement ids. Every approvement definition for selected employee except current one """
        for approvement in self.browse(cr, uid, ids, context=context):
            approver_id = approvement.approver_id.id
        approvement_ids = self.pool.get('hrself.backup.approvement').search(cr, uid, ['&', ('id','!=',ids), ('approver_id','=',approver_id), ], context=context)
        return approvement_ids

hrself_backup_approvement()

class hrself_department_approvement(osv.osv):
    """Department approvement definition."""

    _name = "hrself.department.approvement"
    _inherit = "hrself.approvement"
    _description = "Department approvement definition"

    _columns = {
        'department_id': fields.many2one('hr.department', 'Department', required=True),
    }

    def _get_approvement_ids(self, cr, uid, ids, context=None):
        """ Get approvement ids. Every approvement definition for selected employee except current one """
        for approvement in self.browse(cr, uid, ids, context=context):
            department_id = approvement.department_id.id
        approvement_ids = self.pool.get('hrself.department.approvement').search(cr, uid, ['&', ('id','!=',ids), ('department_id','=',department_id), ], context=context)
        return approvement_ids

hrself_department_approvement()
